@@ -1414,7 +1414,7 @@ test!(
         chain!(
             es2022(es2022::Config { loose: false }),
             es2021(),
-            es2018(),
+            es2018(Default::default()),
             es2017(),
             es2016(),
             es2015::<SingleThreadedComments>(mark, None, Default::default()),
